@@ -1,18 +1,18 @@
 """ Module defining the Django auth backend class for the Keystone API. """
 
-import hashlib
 import logging
 
 from django.conf import settings
 from django.utils.translation import ugettext_lazy as _
 
-from keystoneclient.v2_0 import client as keystone_client
 from keystoneclient import exceptions as keystone_exceptions
-from keystoneclient.v2_0.tokens import Token, TokenManager
 
 from .exceptions import KeystoneAuthException
 from .user import create_user_from_token
-from .utils import check_token_expiration, is_ans1_token
+from .user import Token
+from .utils import check_token_expiration
+from .utils import get_keystone_client
+from .utils import get_keystone_version
 
 
 LOG = logging.getLogger(__name__)
@@ -22,11 +22,12 @@ KEYSTONE_CLIENT_ATTR = "_keystoneclient"
 
 
 class KeystoneBackend(object):
+    """Django authentication backend class for use with
+      ``django.contrib.auth``.
     """
-    Django authentication backend class for use with ``django.contrib.auth``.
-    """
-    def check_auth_expiry(self, token):
-        if not check_token_expiration(token):
+
+    def check_auth_expiry(self, auth_ref):
+        if not check_token_expiration(auth_ref):
             msg = _("The authentication token issued by the Identity service "
                     "has expired.")
             LOG.warning("The authentication token issued by the Identity "
@@ -37,41 +38,41 @@ class KeystoneBackend(object):
         return True
 
     def get_user(self, user_id):
-        """
-        Returns the current user (if authenticated) based on the user ID
+        """Returns the current user (if authenticated) based on the user ID
         and session data.
 
         Note: this required monkey-patching the ``contrib.auth`` middleware
         to make the ``request`` object available to the auth backend class.
         """
         if user_id == self.request.session["user_id"]:
-            token = Token(TokenManager(None),
-                          self.request.session['token'],
-                          loaded=True)
+            token = self.request.session['token']
             endpoint = self.request.session['region_endpoint']
             services_region = self.request.session['services_region']
-            return create_user_from_token(self.request, token, endpoint,
+            user = create_user_from_token(self.request, token, endpoint,
                                           services_region)
+            return user
         else:
             return None
 
     def authenticate(self, request=None, username=None, password=None,
-                     tenant=None, auth_url=None):
-        """ Authenticates a user via the Keystone Identity API. """
+                     user_domain_name=None, auth_url=None):
+        """Authenticates a user via the Keystone Identity API. """
         LOG.debug('Beginning user authentication for user "%s".' % username)
 
         insecure = getattr(settings, 'OPENSTACK_SSL_NO_VERIFY', False)
 
+        keystone_client = get_keystone_client()
         try:
-            client = keystone_client.Client(username=username,
-                                            password=password,
-                                            tenant_id=tenant,
-                                            auth_url=auth_url,
-                                            insecure=insecure)
-            unscoped_token_data = {"token": client.service_catalog.get_token()}
-            unscoped_token = Token(TokenManager(None),
-                                   unscoped_token_data,
-                                   loaded=True)
+            client = keystone_client.Client(
+                user_domain_name=user_domain_name,
+                username=username,
+                password=password,
+                auth_url=auth_url,
+                insecure=insecure,
+                debug=settings.DEBUG)
+
+            unscoped_auth_ref = client.auth_ref
+            unscoped_token = Token(auth_ref=unscoped_auth_ref)
         except (keystone_exceptions.Unauthorized,
                 keystone_exceptions.Forbidden,
                 keystone_exceptions.NotFound) as exc:
@@ -85,54 +86,60 @@ class KeystoneBackend(object):
             LOG.debug(exc.message)
             raise KeystoneAuthException(msg)
 
-        # Check expiry for our unscoped token.
-        self.check_auth_expiry(unscoped_token)
+        # Check expiry for our unscoped auth ref.
+        self.check_auth_expiry(unscoped_auth_ref)
 
-        # FIXME: Log in to default tenant when the Keystone API returns it...
-        # For now we list all the user's tenants and iterate through.
-        try:
-            tenants = client.tenants.list()
-        except (keystone_exceptions.ClientException,
-                keystone_exceptions.AuthorizationFailure):
-            msg = _('Unable to retrieve authorized projects.')
-            raise KeystoneAuthException(msg)
-
-        # Abort if there are no tenants for this user
-        if not tenants:
-            msg = _('You are not authorized for any projects.')
-            raise KeystoneAuthException(msg)
-
-        while tenants:
-            tenant = tenants.pop()
+        # Check if token is automatically scoped to default_project
+        if unscoped_auth_ref.project_scoped:
+            auth_ref = unscoped_auth_ref
+        else:
+            # For now we list all the user's projects and iterate through.
             try:
-                client = keystone_client.Client(tenant_id=tenant.id,
-                                                token=unscoped_token.id,
-                                                auth_url=auth_url,
-                                                insecure=insecure)
-                token = client.tokens.authenticate(username=username,
-                                                   token=unscoped_token.id,
-                                                   tenant_id=tenant.id)
-                break
+                if get_keystone_version() < 3:
+                    projects = client.tenants.list()
+                else:
+                    client.management_url = auth_url
+                    projects = client.projects.list(
+                        user=unscoped_auth_ref.user_id)
             except (keystone_exceptions.ClientException,
-                    keystone_exceptions.AuthorizationFailure):
-                token = None
+                    keystone_exceptions.AuthorizationFailure) as exc:
+                msg = _('Unable to retrieve authorized projects.')
+                raise KeystoneAuthException(msg)
 
-        if token is None:
-            msg = _("Unable to authenticate to any available projects.")
-            raise KeystoneAuthException(msg)
+            # Abort if there are no projects for this user
+            if not projects:
+                msg = _('You are not authorized for any projects.')
+                raise KeystoneAuthException(msg)
+
+            while projects:
+                project = projects.pop()
+                try:
+                    client = keystone_client.Client(
+                        tenant_id=project.id,
+                        token=unscoped_auth_ref.auth_token,
+                        auth_url=auth_url,
+                        insecure=insecure,
+                        debug=settings.DEBUG)
+                    auth_ref = client.auth_ref
+                    break
+                except (keystone_exceptions.ClientException,
+                        keystone_exceptions.AuthorizationFailure):
+                    auth_ref = None
+
+                if auth_ref is None:
+                    msg = _("Unable to authenticate to any available"
+                            " projects.")
+                    raise KeystoneAuthException(msg)
 
         # Check expiry for our new scoped token.
-        self.check_auth_expiry(token)
+        self.check_auth_expiry(auth_ref)
 
         # If we made it here we succeeded. Create our User!
         user = create_user_from_token(request,
-                                      token,
+                                      Token(auth_ref),
                                       client.service_catalog.url_for())
 
         if request is not None:
-            if is_ans1_token(unscoped_token.id):
-                hashed_token = hashlib.md5(unscoped_token.id).hexdigest()
-                unscoped_token._info['token']['id'] = hashed_token
             request.session['unscoped_token'] = unscoped_token.id
             request.user = user
 
@@ -143,15 +150,17 @@ class KeystoneBackend(object):
         return user
 
     def get_group_permissions(self, user, obj=None):
-        """ Returns an empty set since Keystone doesn't support "groups". """
+        """Returns an empty set since Keystone doesn't support "groups". """
+        # Keystone V3 added "groups". The Auth token response includes the
+        # roles from the user's Group assignment. It should be fine just
+        # returning an empty set here.
         return set()
 
     def get_all_permissions(self, user, obj=None):
-        """
-        Returns a set of permission strings that this user has through his/her
-        Keystone "roles".
+        """Returns a set of permission strings that this user has through
+           his/her Keystone "roles".
 
-        The permissions are returned as ``"openstack.{{ role.name }}"``.
+          The permissions are returned as ``"openstack.{{ role.name }}"``.
         """
         if user.is_anonymous() or obj is not None:
             return set()
@@ -163,16 +172,15 @@ class KeystoneBackend(object):
         return role_perms | service_perms
 
     def has_perm(self, user, perm, obj=None):
-        """ Returns True if the given user has the specified permission. """
+        """Returns True if the given user has the specified permission. """
         if not user.is_active:
             return False
         return perm in self.get_all_permissions(user, obj)
 
     def has_module_perms(self, user, app_label):
-        """
-        Returns True if user has any permissions in the given app_label.
+        """Returns True if user has any permissions in the given app_label.
 
-        Currently this matches for the app_label ``"openstack"``.
+           Currently this matches for the app_label ``"openstack"``.
         """
         if not user.is_active:
             return False
